// -*- mode: js2; indent-tabs-mode: nil; js2-basic-offset: 4 -*-
/*

    Copyright (C) 2012-2017  Dieter Schärf <ds@dschaerf.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.

 */

const Lang = imports.lang;

const Gtk = imports.gi.Gtk;
const GObject = imports.gi.GObject;
const Soup = imports.gi.Soup;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;

const Gettext = imports.gettext;
const _ = Gettext.domain(Convenience.CURRENT_GETTEXT_DOMAIN).gettext;

// Soup session (see https://bugzilla.gnome.org/show_bug.cgi?id=661323#c64) (Simon Legner)
const _httpSession = new Soup.SessionAsync();
Soup.Session.prototype.add_feature.call(_httpSession, new Soup.ProxyResolverDefault());

const KEY_LOCATION = 'location';
const KEY_REFRESH_INTERVAL = 'refresh-interval';
const KEY_SHOW_WEATHER_FORECAST = 'show-weather-forecast';
const KEY_WEATHER_ICONS = 'weather-icons';
const KEY_WEATHER_LOCATIONS = 'weather-locations';
const KEY_WEATHER_UNITS = 'weather-units';
const KEY_WIND_DIRECTION = 'wind-direction';

const SETTINGS_GLOBAL = [
    {   schema:  Convenience.CURRENT_SETTINGS_SCHEMA,
        type:    'combo',
        key:     KEY_REFRESH_INTERVAL,
        label:   _("Refresh interval"),
        tooltip: _("Select the desired refresh interval."),
        values:  { '300':  _("5 Minutes"),
                   '600':  _("10 Minutes"),
                   '900':  _("15 Minutes"),
                   '1800': _("30 Minutes"),
                   '3600': _("1 Hour"),
                   '7200': _("2 Hours") }
    },
    {   schema:  Convenience.CURRENT_SETTINGS_SCHEMA,
        type:    'tree',
        key1:    KEY_WEATHER_LOCATIONS,
        key2:    KEY_LOCATION,
        label1:  _("Location"),
        label2:  _("WOEID"),
        buttons: { 'add': _("Add Location"),
                   'del': _("Remove Location") },
        dialog:  '_searchLocation'
    }
];

const SETTINGS_DISPLAY = [
    {   schema:  Convenience.CURRENT_SETTINGS_SCHEMA,
        type:    'combo',
        key:     KEY_WEATHER_UNITS,
        label:   _("System of Units"),
        tooltip: _("Select the system of units for weather."),
        values:  { 'default': _("Anglo American"),
                   'metric':  _("Metric") }
    },
    {   schema:  Convenience.CURRENT_SETTINGS_SCHEMA,
        type:    'combo',
        key:     KEY_WEATHER_ICONS,
        label:   _("Weather Icons"),
        tooltip: _("Select the desired icons for Weather."),
        values:  { 'default':  _("Default"),
                   'system':   _("System icons") }
    },
    {   schema:  Convenience.CURRENT_SETTINGS_SCHEMA,
        type:    'combo',
        key:     KEY_WIND_DIRECTION,
        label:   _("Wind Direction"),
        tooltip: _("Select the display mode of wind direction."),
        values:  { 'arrows':  _("Arrows"),
                   'letters': _("Letters") }
    },
    {   schema:  Convenience.CURRENT_SETTINGS_SCHEMA,
        type:    'switcher',
        key:     KEY_SHOW_WEATHER_FORECAST,
        label:   _("Show Weather Forecast"),
        tooltip: _("If true, display Weather Forecast.")
    }
];

const WeatherPrefsWidget = new GObject.Class({
    Name: 'Weather.Prefs.Widget',
    GTypeName: 'WeatherPrefsWidget',
    Extends: Gtk.Notebook,

    _init: function(params) {
        this.parent(params);
        this._createPrefsWidget();
    },

    _createPrefsWidget: function() {
        let box;
        box = new Gtk.Box({ orientation: Gtk.Orientation.VERTICAL, margin: 15, spacing: 15 });
        for (i in SETTINGS_GLOBAL) {
            box.add(this._createWidget(SETTINGS_GLOBAL[i]));
        }
        this.append_page(box, new Gtk.Label({ label: _("General") }));
        box = new Gtk.Box({ orientation: Gtk.Orientation.VERTICAL, margin: 15, spacing: 15 });
        for (i in SETTINGS_DISPLAY) {
            box.add(this._createWidget(SETTINGS_DISPLAY[i]));
        }
        this.append_page(box, new Gtk.Label({ label: _("Display") }));
    },

    _createWidget: function(args) {
        let box;
        let settings = Convenience.getSettings(args['schema']);
        let widget;
        if (args['type'] == 'tree') {
            box = new Gtk.Box({ orientation: Gtk.Orientation.VERTICAL });
            widget = new Gtk.Frame();
            box.add(widget);
            let scrolledWindow = new Gtk.ScrolledWindow();
            scrolledWindow.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
            widget.add(scrolledWindow);
            let listStore = new Gtk.ListStore();
            listStore.set_column_types([ GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_STRING ]);
            let treeView = new Gtk.TreeView({ expand: true, model: listStore });
            scrolledWindow.add(treeView);
            treeView.get_selection().set_mode(Gtk.SelectionMode.SINGLE);
            let columnWOEID = new Gtk.TreeViewColumn({ expand: true, sort_column_id: 1, title: _(args['label1']) });
            let textRenderer = new Gtk.CellRendererText();
            columnWOEID.pack_start(textRenderer, true);
            columnWOEID.add_attribute(textRenderer, 'text', 1);
            treeView.insert_column(columnWOEID, 0);
            let columnLocation = new Gtk.TreeViewColumn({ expand: true, sort_column_id: 2, title: _(args['label2']) });
            columnLocation.pack_start(textRenderer, true);
            columnLocation.add_attribute(textRenderer, 'text', 2);
            treeView.insert_column(columnLocation, 1);
            if (args['buttons']) {
                let toolbar = new Gtk.Toolbar();
                toolbar.get_style_context().add_class(Gtk.STYLE_CLASS_INLINE_TOOLBAR);
                box.add(toolbar);
                if (args['buttons']['add']) {
                    let buttonADD = new Gtk.ToolButton({ icon_name: 'list-add', label: _(args['buttons']['add']), is_important: true });
                    buttonADD.connect('clicked', Lang.bind (this, function() {
                        this[args['dialog']](settings, args['key1']);
                    }));
                    toolbar.add(buttonADD);
                }
                if (args['buttons']['del']) {
                    let buttonDEL = new Gtk.ToolButton({ icon_name: 'list-remove', label: _(args['buttons']['del']), is_important: true });
                    buttonDEL.connect('clicked', function() {
                        strArray = settings.get_strv(args['key1']);
                        let [ any, model, iter ] = treeView.get_selection().get_selected();
                        if (any && strArray.length > 1) {
                            let item = listStore.get_value(iter, 0);
                            let index = strArray.map(function(i) {
                                return i;
                            }).indexOf(item);
                            if (index < 0) return;
                            strArray.splice(index, 1);
                            settings.set_strv(args['key1'], strArray);
                        }
                    });
                    toolbar.add(buttonDEL);
                }
            }
            let strArray = [];
            let treeViewSelection = treeView.get_selection();
            let updateListStore = function() {
                strArray = settings.get_strv(args['key1']);
                listStore.clear();
                for (let i = 0; i < strArray.length; i++) {
                    let item = strArray[i].split('>');
                    listStore.set(listStore.append(), [ 0, 1, 2 ], [ strArray[i], item[1], item[0] ]);
                }
                treeViewSelection.select_path(Gtk.TreePath.new_from_string(String(settings.get_int(args['key2']))));
            }
            updateListStore();
            treeViewSelection.connect('changed', function() {
                strArray = settings.get_strv(args['key1']);
                let [ any, model, iter ] = treeView.get_selection().get_selected();
                if (any) {
                    let item = listStore.get_value(iter, 0);
                    let index = strArray.map(function(i) {
                        return i;
                    }).indexOf(item);
                    if (index < 0 || index == settings.get_int(args['key2'])) return;
                    settings.set_int(args['key2'], parseInt(index));
                }
            });
            settings.connect('changed::'+args['key1'], updateListStore);
            settings.connect('changed::'+args['key2'], function() {
                treeViewSelection.select_path(Gtk.TreePath.new_from_string(String(settings.get_int(args['key2']))));
            });
        } else {
            box = new Gtk.Box({ orientation: Gtk.Orientation.HORIZONTAL });
            box.pack_start(new Gtk.Label({ label: _(args['label']), xalign: 0 }), true, true, 0);
            switch (args['type']) {
                case 'combo':
                    widget = new Gtk.ComboBoxText();
                    for (let i in args['values']) {
                        widget.append(i, _(args['values'][i]));
                    }
                    widget.set_active_id(settings.get_string(args['key']));
                    widget.width_request = 200;
                    widget.connect('changed', function() {
                        settings.set_string(args['key'], widget.get_active_id());
                    });
                    settings.connect('changed::'+args['key'], function() {
                        widget.set_active_id(settings.get_string(args['key']));
                    });
                    break;
                case 'switcher':
                    widget = new Gtk.Switch({ active: settings.get_boolean(args['key']) });
                    widget.connect('notify::active', function() {
                        settings.set_boolean(args['key'], widget.get_active());
                    });
                    settings.connect('changed::'+args['key'], function() {
                        widget.set_active(settings.get_boolean(args['key']));
                    });
                    break;
            }
            if (args['tooltip']) widget.set_tooltip_text(_(args['tooltip']));
            box.pack_end(widget, false, false, 0);
        }
        return box;
    },

    _loadWeather:function(query, callback) {
        let message = Soup.Message.new('GET', encodeURI('https://query.yahooapis.com/v1/public/yql?q='+query.toLowerCase()+'&format=json'));
        _httpSession.queue_message(message, Lang.bind(this, function(_httpSession, message) {
            try {
                callback.call(this, !message.response_body.data ? null : JSON.parse(message.response_body.data));
            } catch (e) {
                callback.call(this, null);
            }
        }));
    },

    _searchLocation: function(schema, key) {
        let dialog;
        if (Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8) || Convenience.checkGSV(3, 10)) {
            dialog = new Gtk.Dialog({ title: _("Add Location"), modal: true, resizable: false });
        } else {
            dialog = new Gtk.Dialog({ title: _("Add Location"), modal: true, resizable: false, transient_for: this.get_toplevel(), use_header_bar: true });
        }
        dialog.add_button(_("Cancel"), Gtk.ResponseType.CANCEL);
        let buttonOK = dialog.add_button(_("OK"), Gtk.ResponseType.OK);
        buttonOK.sensitive = 0;
        dialog.set_default_response(Gtk.ResponseType.OK);
        let grid = new Gtk.Grid({ margin: 15 });
        dialog.get_content_area().add(grid);
        grid.attach(new Gtk.Label({ label: _("Location:"), xalign: 0, width_request: 130 }), 0, 0, 1, 1);
        let entry = new Gtk.Entry({ activates_default: true, width_request: 300 });
        grid.attach(entry, 1, 0, 1, 1);
        let entryCompletion = new Gtk.EntryCompletion();
        entry.set_completion(entryCompletion);
        let completionModel = Gtk.ListStore.new([ GObject.TYPE_STRING ]);
        entryCompletion.set_model(completionModel);
        entryCompletion.set_text_column(0);
        entryCompletion.set_popup_single_match(true);
        entryCompletion.set_minimum_key_length(1);
        entryCompletion.set_match_func(function(entryCompletion, key, iter) {
            if (iter && completionModel.get_value(iter, 0)) return true;
            return false;
        });
        let checkLocation = Lang.bind(this, function(location) {
            buttonOK.sensitive = 0;
            if (location.search(/\[/) == -1 || location.search(/\]/) == -1) return false;
            let woeid = location.split(/\[/)[1].split(/\]/)[0];
            if (!woeid) return false;
            this._loadWeather('SELECT * FROM geo.places WHERE woeid="'+woeid+'"', function(json) {
                if (typeof arguments[0].query == 'undefined') return;
                if (Number(arguments[0].query.count) < 1) return;
                let city = arguments[0].query.results.place;
                city = city.woeid+'>'+city.name+', '+city.admin1.content+' ('+city.country.code+')';
                let strArray = schema.get_strv(key);
                for (let i = 0; i < strArray.length; i++) {
                    if (strArray[i] == city) return;
                }
                buttonOK.sensitive = 1;
            });
            return true;
        });
        entry.connect('changed', Lang.bind(this, function() {
            let location = entry.get_text();
            if (!checkLocation(location)) this._loadWeather('SELECT * FROM geo.places WHERE text="*'+location+'*" OR woeid="'+location+'"', function(json) {
                if (arguments[0]) {
                    let city = arguments[0].query;
                    if (Number(city.count) > 0) {
                        city = city.results.place;
                        completionModel.clear();
                        let cityEntrys = {};
                        for (let i in city) {
                            let cityCurrent = city[i].name+', '+city[i].admin1.content+' ('+city[i].country.code+')'+' ['+city[i].woeid+']';
                            if (cityEntrys[cityCurrent]) continue;
                            completionModel.set_value(completionModel.append(), 0, cityCurrent);
                            cityEntrys[cityCurrent] = true;
                        }
                        entryCompletion.complete();
                    }
                }
            });
        }));
        dialog.connect('response', Lang.bind(this, function(dialog, response_id) {
            if (response_id == Gtk.ResponseType.OK) {
                if (entry.get_text().search(/\[/) == -1 || entry.get_text().search(/\]/) == -1) return;
                let woeid = entry.get_text().split(/\[/)[1].split(/\]/)[0];
                if (!woeid) return;
                this._loadWeather('SELECT * FROM geo.places WHERE woeid="'+woeid+'"', function(json) {
                    if (typeof arguments[0].query == 'undefined') return;
                    if (Number(arguments[0].query.count) < 1) return;
                    let city = arguments[0].query.results.place;
                    let strArray = schema.get_strv(key);
                    strArray.push(city.woeid+'>'+city.name+', '+city.admin1.content+' ('+city.country.code+')');
                    strArray.sort(function(a, b) {
                        return Convenience.getSorted(a.split('>')[1], b.split('>')[1]);
                    });
                    schema.set_strv(key, strArray);
                    dialog.destroy();
                });
            }
            dialog.destroy();
        }));
        dialog.show_all();
    }
});

function init() {
    Convenience.initTranslations();
}

function buildPrefsWidget() {
    let widget = new WeatherPrefsWidget({
        scrollable: true,
        margin: 15,
        width_request: 720,
        height_request: 480
    });
    widget.show_all();
    return widget;
}
