// -*- mode: js2; indent-tabs-mode: nil; js2-basic-offset: 4 -*-
/*

    Copyright (C) 2012-2017  Dieter Schärf <ds@dschaerf.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.

 */

const Lang = imports.lang;
const Signals = imports.signals;

const Clutter = imports.gi.Clutter;
const GLib = imports.gi.GLib;
const Gtk = imports.gi.Gtk;
const Shell = imports.gi.Shell;
const St = imports.gi.St;

const Params = imports.misc.params;

const Main = imports.ui.main;
const Panel = imports.ui.panel;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;

const Gettext = imports.gettext;
const _ = Gettext.domain(Convenience.CURRENT_GETTEXT_DOMAIN).gettext;

/**
 *  HelpFunctions
 */
function checkApplication(application) {
    try {
        return Shell.AppSystem.get_default().lookup_setting(application);
    } catch (err) {
        return Shell.AppSystem.get_default().lookup_app(application);
    }
}

function getStringCut(text, strLength) {
    strLength = strLength || 50;
    if (text.length <= strLength) return text;
    return text.substr(0, strLength - 3)+'...';
}

function getStringCutOut(text, strLength) {
    strLength = strLength || 50;
    if (text.length <= strLength) return text;
    strLength = Math.round((strLength - 3) / 2);
    return text.substr(0, strLength - 1)+'...'+text.substr(text.length - strLength, strLength);
}

function getStringLineBreak(text, strLength) {
    let start = strLength || 66;
    for (let i = start; i > start - 50; i--) {
        if (text.substr(i, 1) == (' ' || '-')) {
            text = text.substr(0, i + 1)+'\n'+text.substr(i + 1, text.length - i);
            i = start = i + 55;
            if (i > text.length) break;
        }
    }
    return text;
}

/**
 *  Panel
 */
const PANEL_BOX = {
    left:   Main.panel._leftBox,
    center: Main.panel._centerBox,
    right:  Main.panel._rightBox
};

const PANEL_CORNER = {
    left:  Main.panel._leftCorner,
    right: Main.panel._rightCorner
};

const PANEL_CORNER_SET_STYLE_PARENT = Panel.PanelCorner.prototype.setStyleParent ? true : false;

function addToStatusArea(role, indicator, position, box) {
    if (Convenience.checkGSV(3, 4)) {
        if (Main.panel._statusArea[role]) throw new Error(_("Extension point conflict: there is already a status indicator for role '%s'").format(role));
        if (!(indicator instanceof PanelMenu.Button)) throw new TypeError(_("Status indicator must be an instance of PanelMenu.Button"));
        PANEL_BOX[box || 'right'].insert_child_at_index(indicator.actor, position || 0);
        if (indicator.menu) Main.panel._menus.addMenu(indicator.menu);
        Main.panel._statusArea[role] = indicator;
        indicator.actor._rolePosition = position || 0;
        let destroyId = indicator.connect('destroy', Lang.bind(Main.panel, function(emitter) {
            delete Main.panel._statusArea[role];
            emitter.disconnect(destroyId);
        }));
    } else {
        Main.panel.addToStatusArea(role, indicator, position, box);
    }
    return indicator;
}

function getChildren(box) {
    return PANEL_BOX[box].get_children();
}

function getNumChildren(box) {
    return PANEL_BOX[box].get_children().length;
}

function updatePanelCorner(indicator, box) {
    if (!PANEL_CORNER_SET_STYLE_PARENT) return;
    indicator.container.visible = indicator.actor.visible;
    if (Main.panel.actor.get_text_direction() == Clutter.TextDirection.RTL) {
        PANEL_CORNER[box].setStyleParent(box == 'left' ? PANEL_BOX['right'] : PANEL_BOX['left']);
    } else {
        PANEL_CORNER[box].setStyleParent(box == 'left' ? PANEL_BOX['left'] : PANEL_BOX['right']);
    }
}

/**
 *  PanelMenu
 */
const PANELMENU_BUTTON_OFF = 0;
const PANELMENU_BUTTON_ICON_AND_TEXT = 1;
const PANELMENU_BUTTON_ICON = 2;
const PANELMENU_BUTTON_TEXT = 3;

const Button = new Lang.Class({
    Name: 'Button',
    Extends: PanelMenu.Button,

    _init: function(menuAlignment, nameText, params) {
        params = Params.parse(params, { dontCreateMenu: false, icon_name: null, text: null }, true);
        if (menuAlignment != 0 && menuAlignment != 1) {
            menuAlignment = menuAlignment || 0.25;
            if (Clutter.get_default_text_direction() == Clutter.TextDirection.RTL) menuAlignment = 1.0 - menuAlignment;
        }
        this.parent(menuAlignment, nameText || '', params.dontCreateMenu);
        if (!this.accessible_name) this.accessible_name = nameText;
        if (this.menu == null) this.menu = new PopupDummyMenu(this.actor);
        if (Convenience.checkGSV(3, 4)) {
            this._hbox = new St.BoxLayout({ style_class: 'paneltools-panel-status-button-box' });
        } else if (Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8)) {
            this._hbox = new St.BoxLayout({ style_class: 'panel-status-button-box' });
        } else {
            this._hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
        }
        this.actor.add_child(this._hbox);
        if (params.icon_name) {
            this.icon = new St.Icon({ icon_name: params.icon_name, style_class: 'system-status-icon' });
            this._hbox.add(this.icon);
        }
        if (params.text) {
            if (!this.label) this.label = new St.Label({ text: params.text, y_align: Clutter.ActorAlign.CENTER });
            this._hbox.add(this.label);
            this.actor.label_actor = this.label;
        }
        if (!params.dontCreateMenu && PopupMenu.arrowIcon) this._hbox.add(PopupMenu.arrowIcon(St.Side.BOTTOM));
        if (!params.dontCreateMenu && PopupMenu.unicodeArrow) this._hbox.add(PopupMenu.unicodeArrow(St.Side.BOTTOM));
    },

    updateButton: function(showButton) {
        this.updateVisibility(showButton != PANELMENU_BUTTON_OFF);
        if (this.icon) this.icon.visible = showButton == PANELMENU_BUTTON_ICON_AND_TEXT || showButton == PANELMENU_BUTTON_ICON;
        if (this.label) this.label.visible = showButton == PANELMENU_BUTTON_ICON_AND_TEXT || showButton == PANELMENU_BUTTON_TEXT;
    },

    updateVisibility: function(visible) {
        this.actor.visible = visible;
    }
});

/**
 *  PopupMenu
 */
const KEY_MENUS_HAVE_ICONS = 'menus-have-icons';
const KEY_MENUS_ICON_SIZE = 'menus-icon-size';

// PanelManager@dschaerf required to changed the settings for icon size and visibility
const PANELMANAGER_SETTINGS_SCHEMA = 'de.dschaerf.extensions.panelManager';

const POPUPMENU_MENUS_ICON_SIZE = 18;
const POPUPMENU_MENUS_HAVE_ICONS = true;

const PopupMenuItem = new Lang.Class({
    Name: 'PopupMenuItem',
    Extends: PopupMenu.PopupBaseMenuItem,

    _init: function (text, wantIcon, params) {
        this.parent(params);
        if (Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8)) {
            this._hbox = new St.BoxLayout({ style_class: 'paneltools-popup-menu-box' });
            this.addActor(this._hbox, { expand: true });
        }
        if (wantIcon) {
            this.icon = this._application ? new St.Bin({ style_class: 'popup-menu-icon' }) : new St.Icon({ style_class: 'popup-menu-icon' });
            (this._hbox || this.actor).add(this.icon);
            this._settings = Convenience.getSettings(PANELMANAGER_SETTINGS_SCHEMA, true);
            if (this._settings) this._connectSettings = [
                this._settings.connect('changed::'+KEY_MENUS_HAVE_ICONS, Lang.bind(this, this._setIconVisible)),
                this._settings.connect('changed::'+KEY_MENUS_ICON_SIZE, Lang.bind(this, this._setIconSize))
            ];
            this._setIconSize();
            this._setIconVisible();
        }
        this.label = new St.Label({ text: text, y_align: Clutter.ActorAlign.CENTER });
        this.actor.label_actor = this.label;
        (this._hbox || this.actor).add(this.label);
        if (Convenience.checkGSV(3, 4)) {
            this.status = new St.Label({ style_class: 'popup-inactive-menu-item', y_align: Clutter.ActorAlign.CENTER });
        } else {
            this.status = new St.Label({ style_class: 'popup-status-menu-item', y_align: Clutter.ActorAlign.CENTER });
        }
        (this._hbox || this.actor).add(new St.Bin(), { expand: true });
        (this._hbox || this.actor).add(this.status);
        if (this._connectSettings) this.actor.connect('destroy', Lang.bind(this, function() {
            for each(let connection in this._connectSettings) {
                this._settings.disconnect(connection);
            }
        }));
    },

    _setIconSize: function() {
        let iconSize = this._settings ? this._settings.get_int(KEY_MENUS_ICON_SIZE) : POPUPMENU_MENUS_ICON_SIZE;
        if (this._application) {
            this.icon.child = this._application.create_icon_texture(iconSize);
        } else {
            this.icon.icon_size = iconSize;
        }
        this.icon.width = iconSize;
        if (this.button) this.button.icon.icon_size = iconSize - 2;
    },

    _setIconVisible: function() {
        this.icon.visible = this._settings ? this._settings.get_boolean(KEY_MENUS_HAVE_ICONS) : POPUPMENU_MENUS_HAVE_ICONS;
    },

    addButton: function(iconName, callback) {
        this.button = new St.Button({ style_class: 'paneltools-popup-menu-button' });
        this.button.icon = new St.Icon({ icon_name: iconName })
        this.button.child = this.button.icon;
        (this._hbox || this.actor).add(this.button, { align: St.Align.END });
        this.button.connect('clicked', Lang.bind(this, function(event) {
            callback(event);
        }));
        this._setIconSize();
    },

    setIconName: function(iconName) {
        this.icon.icon_name = iconName;
        if (St.IconType && iconName.indexOf('-symbolic') == -1) this.icon.icon_type = St.IconType.FULLCOLOR;
    },

    setSelected: function(value) {
        if (Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8)) {
            this.setShowDot(value);
        } else {
            this.setOrnament(value ? PopupMenu.Ornament.CHECK : PopupMenu.Ornament.NONE);
        }
    },

    setStatus: function(text) {
        this.status.set_text(text);
    }
});

const PopupActionMenuItem = new Lang.Class({
    Name: 'PopupActionMenuItem',
    Extends: PopupMenu.PopupBaseMenuItem,

    _init: function (params) {
        params = Params.parse(params, { span: 1 }, true);
        if (Convenience.checkGSV(3, 4)) {
            this.parent({ reactive: false });
        } else {
            this.parent({ reactive: false, can_focus: false });
        }
        if (Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8)) {
            this._hbox = new St.BoxLayout({ style_class: 'paneltools-popup-menu-box' });
            this.addActor(this._hbox, { expand: true, span: params.span });
        }
    },

    addActionButton: function(icon, accessibleName) {
        let actionButton;
        if (Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8)) {
            actionButton = new St.Button({ reactive: true, can_focus: true, track_hover: true, accessible_name: accessibleName, style_class: 'paneltools-system-menu-action' });
        } else {
            actionButton = new St.Button({ reactive: true, can_focus: true, track_hover: true, accessible_name: accessibleName, style_class: 'system-menu-action' });
        }
        if (typeof icon == 'string') {
            actionButton.child = new St.Icon({ icon_name: icon });
        } else {
            actionButton.child = icon;
        }
        (this._hbox || this.actor).add(actionButton, { expand: true, x_fill: false });
        return actionButton;
    }
});

const PopupApplicationMenuItem = new Lang.Class({
    Name: 'PopupApplicationMenuItem',
    Extends: PopupMenuItem,

    _init: function(app, wantIcon, params) {
        this._application = app;
        this.parent(app.get_name(), wantIcon, params);
        if (wantIcon) {
            this._setIconSize();
            this._setIconVisible();
        }
        this.connect('activate', Lang.bind(this, function() {
            Main.overview.hide();
            app.activate(-1);
        }));
    }
});

//  Only for GNOME Shell 3.4 and 3.6 for PanelMenu.Button without menu
if (Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6)) {
const PopupDummyMenu = new Lang.Class({
    Name: 'PopupDummyMenu',

    _init: function(sourceActor) {
        this.sourceActor = sourceActor;
        this.actor = sourceActor;
        this.actor._delegate = this;
    },

    getSensitive: function() {
        return true;
    },

    isChildMenu: function() {
        return false;
    },

    open: function() {
        this.emit('open-state-changed', true);
    },

    close: function() {
        this.emit('open-state-changed', false);
    },

    toggle: function() {},

    destroy: function() {
        this.emit('destroy');
    },
});
Signals.addSignalMethods(PopupDummyMenu.prototype);
}

const PopupSubMenuMenuItem = new Lang.Class({
    Name: 'PopupSubMenuMenuItem',
    Extends: PopupMenu.PopupSubMenuMenuItem,

    _init: function(text, wantIcon, params) {
        params = Params.parse(params, { settings: true, style_class: 'popup-menu-icon', span: 1 }, true);
        if (Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8)) {
            this.parent(text, false);
            this.removeActor(this.label);
            this.removeActor(this._triangle);
            this._hbox = new St.BoxLayout({ style_class: 'paneltools-popup-menu-box' });
            this.addActor(this._hbox, { expand: true, span: params.span });
            if (wantIcon) {
                this.icon = new St.Icon({ style_class: params.style_class });
                this._hbox.add(this.icon);
            }
            this.label = new St.Label({ text: text, y_align: Clutter.ActorAlign.CENTER });
            this.actor.label_actor = this.label;
            this._hbox.add(this.label);
            this._hbox.add(new St.Bin(), { expand: true });
            if (Convenience.checkGSV(3, 4)) {
                this.status = new St.Label({ style_class: 'popup-inactive-menu-item', y_align: Clutter.ActorAlign.CENTER });
            } else {
                this.status = new St.Label({ style_class: 'popup-status-menu-item', y_align: Clutter.ActorAlign.CENTER });
            }
            this._hbox.add(this.status);
            this._hbox.add(this._triangle, { align: St.Align.END });
        } else {
            this.parent(text, wantIcon);
        }
        if (wantIcon && params.settings) {
            this._settings = Convenience.getSettings(PANELMANAGER_SETTINGS_SCHEMA, true);
            if (this._settings) this._connectSettings = [
                this._settings.connect('changed::'+KEY_MENUS_HAVE_ICONS, Lang.bind(this, this._setIconVisible)),
                this._settings.connect('changed::'+KEY_MENUS_ICON_SIZE, Lang.bind(this, this._setIconSize))
            ];
            this._setIconSize();
            this._setIconVisible();
        }
        if (this._connectSettings) this.actor.connect('destroy', Lang.bind(this, function() {
            for each(let connection in this._connectSettings) {
                this._settings.disconnect(connection);
            }
        }));
    },

    _setIconSize: function() {
        this.icon.icon_size = this._settings ? this._settings.get_int(KEY_MENUS_ICON_SIZE) : POPUPMENU_MENUS_ICON_SIZE;
        this.icon.width = this._settings ? this._settings.get_int(KEY_MENUS_ICON_SIZE) : POPUPMENU_MENUS_ICON_SIZE;
    },

    _setIconVisible: function() {
        this.icon.visible = this._settings ? this._settings.get_boolean(KEY_MENUS_HAVE_ICONS) : POPUPMENU_MENUS_HAVE_ICONS;
    },

    setIconName: function(iconName) {
        this.icon.icon_name = iconName;
        if (St.IconType && iconName.indexOf('-symbolic') == -1) this.icon.icon_type = St.IconType.FULLCOLOR;
    },

    setStatus: function(text) {
        this.status.set_text(text);
    }
});
